# -*- coding: utf-8 -*-
'''
Util functions for the NXOS API modules.
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python std lib
import json
import logging

# Import salt libs
import salt.utils.http
from salt.ext import six
from salt.exceptions import SaltException
try:
    from salt.utils.args import clean_kwargs
except ImportError:
    from salt.utils import clean_kwargs

log = logging.getLogger(__name__)

RPC_INIT_KWARGS = [
    'transport',
    'host',
    'username',
    'password',
    'port',
    'timeout',
    'verify',
    'rpc_version'
]


def _prepare_connection(**nxos_api_kwargs):
    '''
    Prepare the connection with the remote network device, and clean up the key
    value pairs, removing the args used for the connection init.
    '''
    nxos_api_kwargs = clean_kwargs(**nxos_api_kwargs)
    init_kwargs = {}
    # Clean up any arguments that are not required
    for karg, warg in six.iteritems(nxos_api_kwargs):
        if karg in RPC_INIT_KWARGS:
            init_kwargs[karg] = warg
    if 'host' not in init_kwargs:
        init_kwargs['host'] = 'localhost'
    if 'transport' not in init_kwargs:
        init_kwargs['transport'] = 'https'
    if 'port' not in init_kwargs:
        init_kwargs['port'] = 80 if init_kwargs['transport'] == 'http' else 443
    verify = init_kwargs.get('verify', True)
    if isinstance(verify, bool):
        init_kwargs['verify_ssl'] = verify
    else:
        init_kwargs['ca_bundle'] = verify
    if 'rpc_version' not in init_kwargs:
        init_kwargs['rpc_version'] = '2.0'
    if 'timeout' not in init_kwargs:
        init_kwargs['timeout'] = 60
    return init_kwargs


def rpc(commands,
        method='cli',
        **kwargs):
    '''
    Execute an arbitrary RPC request via the Nexus API.

    commands
        The commands to be executed.

    method: ``cli``
        The type of the response, i.e., raw text (``cli_ascii``) or structured
        document (``cli``). Defaults to ``cli`` (structured data).

    transport: ``https``
        Specifies the type of connection transport to use. Valid values for the
        connection are ``http``, and  ``https``.

    host: ``localhost``
        The IP address or DNS host name of the connection device.

    username: ``admin``
        The username to pass to the device to authenticate the NX-API connection.

    password
        The password to pass to the device to authenticate the NX-API connection.

    port
        The TCP port of the endpoint for the NX-API connection. If this keyword is
        not specified, the default value is automatically determined by the
        transport type (``80`` for ``http``, or ``443`` for ``https``).

    timeout: ``60``
        Time in seconds to wait for the device to respond. Default: 60 seconds.

    verify: ``True``
        Either a boolean, in which case it controls whether we verify the NX-API
        TLS certificate, or a string, in which case it must be a path to a CA bundle
        to use. Defaults to ``True``.
    '''
    init_args = _prepare_connection(**kwargs)
    log.error('These are the init args:')
    log.error(init_args)
    url = '{transport}://{host}:{port}/ins'.format(
        transport=init_args['transport'],
        host=init_args['host'],
        port=init_args['port']
    )
    headers = {
        'content-type': 'application/json-rpc'
    }
    payload = []
    if not isinstance(commands, (list, tuple)):
        commands = [commands]
    for index, command in enumerate(commands):
        payload.append({
            'jsonrpc': init_args['rpc_version'],
            'method': method,
            'params': {
                'cmd': command,
                'version': 1
            },
            'id': index + 1
        })
    opts = {
        'http_request_timeout': init_args['timeout']
    }
    response = salt.utils.http.query(url,
                                     method='POST',
                                     opts=opts,
                                     data=json.dumps(payload),
                                     header_dict=headers,
                                     decode=True,
                                     decode_type='json',
                                     **init_args)
    if 'error' in response:
        raise SaltException(response['error'])
    response_list = response['dict']
    if isinstance(response_list, dict):
        response_list = [response_list]
    for index, command in enumerate(commands):
        response_list[index]['command'] = command
    return response_list



# -*- coding: utf-8 -*-
'''
Read in the roster from a flat file using the renderer system
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.loader
import salt.config
from salt.ext import six
from salt.template import compile_template
from salt.roster import get_roster_file

import logging
log = logging.getLogger(__name__)


def targets(tgt, tgt_type='glob', **kwargs):
    '''
    Return the targets from the flat yaml file, checks opts for location but
    defaults to /etc/salt/roster
    '''
    template = get_roster_file(__opts__)

    rend = salt.loader.render(__opts__, {})
    raw = compile_template(template,
                           rend,
                           __opts__['renderer'],
                           __opts__['renderer_blacklist'],
                           __opts__['renderer_whitelist'],
                           mask_value='passw*',
                           **kwargs)
    conditioned_raw = {}
    for minion in raw:
        conditioned_raw[six.text_type(minion)] = salt.config.apply_sdb(raw[minion])
    return __utils__['roster_matcher.targets'](conditioned_raw, tgt, tgt_type, 'ipv4')



# -*- coding: utf-8 -*-
'''
The matcher subsystem needs a function called 'confirm_top', which
takes the data passed to a top file environment and determines if that
data matches this minion.
'''
from __future__ import absolute_import
import logging

import salt.loader

log = logging.getLogger(__file__)


def confirm_top(match, data, nodegroups=None):
    '''
    Takes the data passed to a top file environment and determines if the
    data matches this minion
    '''
    matcher = 'compound'
    if not data:
        log.error('Received bad data when setting the match from the top '
                  'file')
        return False
    for item in data:
        if isinstance(item, dict):
            if 'match' in item:
                matcher = item['match']

    matchers = salt.loader.matchers(__opts__)
    funcname = matcher + '_match.match'
    if matcher == 'nodegroup':
        return matchers[funcname](match, nodegroups)
    else:
        m = matchers[funcname]
        return m(match)
    # except TypeError, KeyError:
    #     log.error('Attempting to match with unknown matcher: %s', matcher)



# -*- coding: utf-8 -*-
'''
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.mock import (
    MagicMock,
    patch,
    mock_open,
    NO_MOCK,
    NO_MOCK_REASON
)

# Import Salt Libs
import salt.modules.apache as apache
from salt.ext.six.moves.urllib.error import URLError  # pylint: disable=import-error,no-name-in-module
from salt.utils.odict import OrderedDict


@skipIf(NO_MOCK, NO_MOCK_REASON)
class ApacheTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for salt.modules.apache
    '''

    def setup_loader_modules(self):
        return {apache: {}}

    # 'version' function tests: 1
    def test_version(self):
        '''
        Test if return server version (``apachectl -v``)
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value="Server version: Apache/2.4.7")
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.version() == 'Apache/2.4.7'

    # 'fullversion' function tests: 1

    def test_fullversion(self):
        '''
        Test if return server version (``apachectl -V``)
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value="Server version: Apache/2.4.7")
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.fullversion() == {'compiled_with': [],
                                                'server_version': 'Apache/2.4.7'}

    # 'modules' function tests: 1

    def test_modules(self):
        '''
        Test if return list of static and shared modules
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value=
                             "unixd_module (static)\n \
                             access_compat_module (shared)")
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.modules() == {'shared': ['access_compat_module'],
                                            'static': ['unixd_module']}

    # 'servermods' function tests: 1

    def test_servermods(self):
        '''
        Test if return list of modules compiled into the server
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value="core.c\nmod_so.c")
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.servermods() == ['core.c', 'mod_so.c']

    # 'directives' function tests: 1

    def test_directives(self):
        '''
        Test if return list of directives
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value="Salt")
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.directives() == {'Salt': ''}

    # 'vhosts' function tests: 1

    def test_vhosts(self):
        '''
        Test if it shows the virtualhost settings
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value='')
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.vhosts() == {}

    # 'signal' function tests: 2

    def test_signal(self):
        '''
        Test if return no signal for httpd
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            mock = MagicMock(return_value='')
            with patch.dict(apache.__salt__, {'cmd.run': mock}):
                assert apache.signal(None) is None

    def test_signal_args(self):
        '''
        Test if return httpd signal to start, restart, or stop.
        '''
        with patch('salt.modules.apache._detect_os',
                   MagicMock(return_value='apachectl')):
            ret = 'Command: "apachectl -k start" completed successfully!'
            mock = MagicMock(return_value={'retcode': 1,
                                           'stderr': '',
                                           'stdout': ''})
            with patch.dict(apache.__salt__, {'cmd.run_all': mock}):
                assert apache.signal('start') == ret

            mock = MagicMock(return_value={'retcode': 1,
                                           'stderr': 'Syntax OK',
                                           'stdout': ''})
            with patch.dict(apache.__salt__, {'cmd.run_all': mock}):
                assert apache.signal('start') == 'Syntax OK'

            mock = MagicMock(return_value={'retcode': 0,
                                           'stderr': 'Syntax OK',
                                           'stdout': ''})
            with patch.dict(apache.__salt__, {'cmd.run_all': mock}):
                assert apache.signal('start') == 'Syntax OK'

            mock = MagicMock(return_value={'retcode': 1,
                                           'stderr': '',
                                           'stdout': 'Salt'})
            with patch.dict(apache.__salt__, {'cmd.run_all': mock}):
                assert apache.signal('start') == 'Salt'

    # 'useradd' function tests: 1

    def test_useradd(self):
        '''
        Test if it add HTTP user using the ``htpasswd`` command
        '''
        mock = MagicMock(return_value=True)
        with patch.dict(apache.__salt__, {'webutil.useradd': mock}):
            assert apache.useradd('htpasswd', 'salt', 'badpassword') is True

    # 'userdel' function tests: 1

    def test_userdel(self):
        '''
        Test if it delete HTTP user using the ``htpasswd`` file
        '''
        mock = MagicMock(return_value=True)
        with patch.dict(apache.__salt__, {'webutil.userdel': mock}):
            assert apache.userdel('htpasswd', 'salt') is True

    # 'server_status' function tests: 2

    def test_server_status(self):
        '''
        Test if return get information from the Apache server-status
        '''
        with patch('salt.modules.apache.server_status', MagicMock(return_value={})):
            mock = MagicMock(return_value='')
            with patch.dict(apache.__salt__, {'config.get': mock}):
                assert apache.server_status() == {}

    def test_server_status_error(self):
        '''
        Test if return get error from the Apache server-status
        '''
        mock = MagicMock(side_effect=URLError('error'))
        with patch.object(apache, '_urlopen', mock):
            mock = MagicMock(return_value='')
            with patch.dict(apache.__salt__, {'config.get': mock}):
                assert apache.server_status() == 'error'

    # 'config' function tests: 1

    def test_config(self):
        '''
        Test if it create VirtualHost configuration files
        '''
        with patch('salt.modules.apache._parse_config',
                   MagicMock(return_value='Listen 22')):
            with patch('salt.utils.files.fopen', mock_open()):
                assert apache.config('/ports.conf', [{'Listen': '22'}]) == 'Listen 22'

    # '_parse_config' function tests: 2

    def test__parse_config_dict(self):
        '''
        Test parsing function which creates configs from dict like (legacy way):
            - VirtualHost:
              this: '*:80'
              ServerName: website.com
              ServerAlias:
                - www
                - dev
              Directory:
                  this: /var/www/vhosts/website.com
                  Order: Deny,Allow
                  Allow from:
                    - 127.0.0.1
                    - 192.168.100.0/24

        '''
        data_in = OrderedDict([
            ('Directory', OrderedDict([
                ('this', '/var/www/vhosts/website.com'),
                ('Order', 'Deny,Allow'),
                ('Allow from', ['127.0.0.1', '192.168.100.0/24'])])),
            ('this', '*:80'),
            ('ServerName', 'website.com'),
            ('ServerAlias', ['www', 'dev'])
        ])
        dataout = '<VirtualHost *:80>\n' \
                  '<Directory /var/www/vhosts/website.com>\n' \
                  'Order Deny,Allow\n' \
                  'Allow from 127.0.0.1\n' \
                  'Allow from 192.168.100.0/24\n\n' \
                  '</Directory>\n\n' \
                  'ServerName website.com\n' \
                  'ServerAlias www\n' \
                  'ServerAlias dev\n\n' \
                  '</VirtualHost>\n'
        # pylint: disable=protected-access
        parse = apache._parse_config(data_in, 'VirtualHost')
        assert parse == dataout

    def test__parse_config_list(self):
        '''
        Test parsing function which creates configs from variable structure (list of dicts or
        list of dicts of dicts/lists) like:
            - VirtualHost:
              - this: '*:80'
              - ServerName: website.com
              - ServerAlias:
                - www
                - dev
              - Directory:
                  this: /var/www/vhosts/website.com
                  Order: Deny,Allow
                  Allow from:
                    - 127.0.0.1
                    - 192.168.100.0/24
              - Directory:
                - this: /var/www/vhosts/website.com/private
                - Order: Deny,Allow
                - Allow from:
                  - 127.0.0.1
                  - 192.168.100.0/24
                - If:
                    this: some condition
                    do: something
        '''
        data_in = [OrderedDict([
            ('ServerName', 'website.com'),
            ('ServerAlias', ['www', 'dev']),
            ('Directory', [OrderedDict([
                ('this', '/var/www/vhosts/website.com/private'),
                ('Order', 'Deny,Allow'),
                ('Allow from', ['127.0.0.1', '192.168.100.0/24']),
                ('If', {'this': 'some condition', 'do': 'something'})
            ])]),
            ('this', '*:80')
        ])]
        dataout = '<VirtualHost *:80>\n' \
                  'ServerName website.com\n' \
                  'ServerAlias www\n' \
                  'ServerAlias dev\n\n' \
                  '<Directory /var/www/vhosts/website.com/private>\n' \
                  'Order Deny,Allow\n' \
                  'Allow from 127.0.0.1\n' \
                  'Allow from 192.168.100.0/24\n\n' \
                  '<If some condition>\n' \
                  'do something\n' \
                  '</If>\n\n' \
                  '</Directory>\n\n' \
                  '</VirtualHost>\n'
        # pylint: disable=protected-access
        parse = apache._parse_config(data_in, 'VirtualHost')
        assert parse == dataout



# -*- coding: utf-8 -*-
'''
This is the default pillar exact matcher.
'''
from __future__ import absolute_import, print_function, unicode_literals

import logging
import salt.utils.data  # pylint: disable=3rd-party-module-not-gated

log = logging.getLogger(__name__)


def match(tgt, delimiter=':'):
    '''
    Reads in the pillar match, no globbing, no PCRE
    '''
    log.debug('pillar target: %s', tgt)
    if delimiter not in tgt:
        log.error('Got insufficient arguments for pillar match '
                  'statement from master')
        return False
    return salt.utils.data.subdict_match(__opts__['pillar'],
                                         tgt,
                                         delimiter=delimiter,
                                         exact_match=True)



# -*- coding: utf-8 -*-
'''
Utility functions for SMB connections

:depends: impacket
'''

from __future__ import absolute_import, print_function, unicode_literals
import socket
import uuid

# Import python libs
import salt.utils.files
import salt.utils.stringutils
import salt.utils.versions
import logging

log = logging.getLogger(__name__)

from salt.exceptions import MissingSmb
try:
    import impacket.smbconnection
    from impacket.smbconnection import SessionError as smbSessionError
    from impacket.smb3 import SessionError as smb3SessionError
    HAS_IMPACKET = True
except ImportError:
    HAS_IMPACKET = False

try:
    from smbprotocol.connection import Connection
    from smbprotocol.session import Session
    from smbprotocol.tree import TreeConnect
    from smbprotocol.open import (
        Open, ImpersonationLevel, FilePipePrinterAccessMask, FileAttributes,
        CreateDisposition, CreateOptions, ShareAccess, DirectoryAccessMask,
        FileInformationClass
    )
    from smbprotocol.create_contexts import (
        CreateContextName,
        SMB2CreateContextRequest, SMB2CreateQueryMaximalAccessRequest
    )
    from smbprotocol.security_descriptor import (
        AccessAllowedAce, AccessMask, AclPacket, SDControl, SIDPacket,
        SMB2CreateSDBuffer
    )
    logging.getLogger('smbprotocol').setLevel(logging.WARNING)
    HAS_SMBPROTOCOL = True
except ImportError:
    HAS_SMBPROTOCOL = False


class SMBProto(object):

    def __init__(self, server, username, password, port=445):
        connection_id = uuid.uuid4()
        addr = socket.gethostbyname(server)
        self.server = server
        connection = Connection(connection_id, addr, port, require_signing=True)
        self.session = Session(connection, username, password, require_encryption=False)

    def connect(self):
        self.connection.connect()
        self.session.connect()

    def close(self):
        self.session.connection.disconnect(True)

    @property
    def connection(self):
        return self.session.connection

    def tree_connect(self, share):
        if share.endswith('$'):
            share = r'\\{}\{}'.format(self.server, share)
        tree = TreeConnect(self.session, share)
        tree.connect()
        return tree

    @staticmethod
    def normalize_filename(file):
        return file.lstrip('\\')

    @classmethod
    def open_file(cls, tree, file):
        file = cls.normalize_filename(file)
        # ensure file is created, get maximal access, and set everybody read access
        max_req = SMB2CreateContextRequest()
        max_req['buffer_name'] = CreateContextName.SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST
        max_req['buffer_data'] = SMB2CreateQueryMaximalAccessRequest()

        # create security buffer that sets the ACL for everyone to have read access
        everyone_sid = SIDPacket()
        everyone_sid.from_string("S-1-1-0")
        ace = AccessAllowedAce()
        ace['mask'] = AccessMask.GENERIC_ALL
        ace['sid'] = everyone_sid
        acl = AclPacket()
        acl['aces'] = [ace]
        sec_desc = SMB2CreateSDBuffer()
        sec_desc['control'].set_flag(SDControl.SELF_RELATIVE)
        sec_desc.set_dacl(acl)
        sd_buffer = SMB2CreateContextRequest()
        sd_buffer['buffer_name'] = CreateContextName.SMB2_CREATE_SD_BUFFER
        sd_buffer['buffer_data'] = sec_desc

        create_contexts = [
            max_req,
            sd_buffer
        ]
        file_open = Open(tree, file)
        open_info = file_open.create(
            ImpersonationLevel.Impersonation,
            FilePipePrinterAccessMask.GENERIC_READ |
            FilePipePrinterAccessMask.GENERIC_WRITE,
            FileAttributes.FILE_ATTRIBUTE_NORMAL,
            ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
            CreateDisposition.FILE_OVERWRITE_IF,
            CreateOptions.FILE_NON_DIRECTORY_FILE,
        )
        return file_open

    @staticmethod
    def open_directory(tree, name, create=False):
        # ensure directory is created
        dir_open = Open(tree, name)
        if create:
            dir_open.create(
                ImpersonationLevel.Impersonation,
                DirectoryAccessMask.GENERIC_READ | DirectoryAccessMask.GENERIC_WRITE,
                FileAttributes.FILE_ATTRIBUTE_DIRECTORY,
                ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
                CreateDisposition.FILE_OPEN_IF,
                CreateOptions.FILE_DIRECTORY_FILE
            )
        return dir_open


class StrHandle(object):
    '''
    Fakes a file handle, so that raw strings may be uploaded instead of having
    to write files first. Used by put_str()
    '''
    def __init__(self, content):
        '''
        Init
        '''
        self.content = content
        self.finished = False

    def string(self, writesize=None):
        '''
        Looks like a file handle
        '''
        if not self.finished:
            self.finished = True
            return self.content
        return ''


def _get_conn_impacket(host=None, username=None, password=None, client_name=None, port=445):
    conn = impacket.smbconnection.SMBConnection(
        remoteName=host,
        remoteHost=host,
        myName=client_name,
    )
    conn.login(user=username, password=password)
    return conn


def _get_conn_smbprotocol(host='', username='', password='', client_name='', port=445):
    conn = SMBProto(host, username, password, port)
    conn.connect()
    return conn


def get_conn(host='', username=None, password=None, port=445):
    '''
    Get an SMB connection
    '''
    if HAS_IMPACKET and not HAS_SMBPROTOCOL:
        salt.utils.versions.warn_until(
            'Sodium',
            'Support of impacket has been depricated and will be '
            'removed in Sodium. Please install smbprotocol instead.'
        )
    if HAS_SMBPROTOCOL:
        log.info('Get connection smbprotocol')
        return _get_conn_smbprotocol(host, username, password, port=port)
    elif HAS_IMPACKET:
        log.info('Get connection impacket')
        return _get_conn_impacket(host, username, password, port=port)
    return False


def _mkdirs_impacket(path, share='C$', conn=None, host=None, username=None, password=None):
    '''
    Recursively create a directory structure on an SMB share

    Paths should be passed in with forward-slash delimiters, and should not
    start with a forward-slash.
    '''
    if conn is None:
        conn = get_conn(host, username, password)

    if conn is False:
        return False

    comps = path.split('/')
    pos = 1
    for comp in comps:
        cwd = '\\'.join(comps[0:pos])
        try:
            conn.listPath(share, cwd)
        except (smbSessionError, smb3SessionError):
            log.exception('Encountered error running conn.listPath')
            conn.createDirectory(share, cwd)
        pos += 1


def _mkdirs_smbprotocol(path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)

    if conn is False:
        return False

    tree = conn.tree_connect(share)
    comps = path.split('/')
    pos = 1
    for comp in comps:
        cwd = '\\'.join(comps[0:pos])
        dir_open = conn.open_directory(tree, cwd, create=True)
        compound_messages = [
            dir_open.query_directory("*",
                FileInformationClass.FILE_NAMES_INFORMATION,
                send=False),
            dir_open.close(False, send=False)
        ]
        requests = conn.session.connection.send_compound([x[0] for x in compound_messages],
                                        conn.session.session_id,
                                        tree.tree_connect_id)
        for i, request in enumerate(requests):
            response = compound_messages[i][1](request)
        pos += 1


def mkdirs(path, share='C$', conn=None, host=None, username=None, password=None):
    if HAS_SMBPROTOCOL:
        return _mkdirs_smbprotocol(path, share, conn=conn, host=host, username=username, password=password)
    elif HAS_IMPACKET:
        return _mkdirs_impacket(path, share, conn=conn, host=host, username=username, password=password)
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _put_str_impacket(content, path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)

    if conn is False:
        return False

    fh_ = StrHandle(content)
    conn.putFile(share, path, fh_.string)


def _put_str_smbprotocol(
        content, path, share='C$', conn=None, host=None, username=None,
        password=None):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    tree = conn.tree_connect(share)
    try:
        file_open = conn.open_file(tree, path)
        file_open.write(salt.utils.stringutils.to_bytes(content), 0)
    finally:
        file_open.close()


def put_str(content, path, share='C$', conn=None, host=None, username=None, password=None):
    '''
    Wrapper around impacket.smbconnection.putFile() that allows a string to be
    uploaded, without first writing it as a local file
    '''
    if HAS_SMBPROTOCOL:
        return _put_str_smbprotocol(
            content, path, share, conn=conn, host=host,
            username=username, password=password
        )
    elif HAS_IMPACKET:
        return _put_str_impacket(
            content, path, share, conn=conn, host=host, username=username, password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _put_file_impacket(local_path, path, share='C$', conn=None, host=None, username=None, password=None):
    '''
    Wrapper around impacket.smbconnection.putFile() that allows a file to be
    uploaded

    Example usage:

        import salt.utils.smb
        smb_conn = salt.utils.smb.get_conn('10.0.0.45', 'vagrant', 'vagrant')
        salt.utils.smb.put_file('/root/test.pdf', 'temp\\myfiles\\test1.pdf', conn=smb_conn)
    '''
    if conn is None:
        conn = get_conn(host, username, password)

    if conn is False:
        return False

    if hasattr(local_path, 'read'):
        conn.putFile(share, path, local_path)
        return
    with salt.utils.files.fopen(local_path, 'rb') as fh_:
        conn.putFile(share, path, fh_.read)


def _put_file_smbprotocol(
        local_path, path, share='C$', conn=None, host=None, username=None,
        password=None, chunk_size=1024 * 1024):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False

    tree = conn.tree_connect(share)
    file_open = conn.open_file(tree, path)
    with salt.utils.files.fopen(local_path, 'rb') as fh_:
        try:
            position = 0
            while True:
                chunk = fh_.read(chunk_size)
                if not chunk:
                    break
                file_open.write(chunk, position)
                position += len(chunk)
        finally:
            file_open.close(False)


def put_file(local_path, path, share='C$', conn=None, host=None, username=None, password=None):
    '''
    Wrapper around impacket.smbconnection.putFile() that allows a file to be
    uploaded

    Example usage:

        import salt.utils.smb
        smb_conn = salt.utils.smb.get_conn('10.0.0.45', 'vagrant', 'vagrant')
        salt.utils.smb.put_file('/root/test.pdf', 'temp\\myfiles\\test1.pdf', conn=smb_conn)
    '''
    if HAS_SMBPROTOCOL:
        return _put_file_smbprotocol(
            local_path, path, share, conn=conn, host=host, username=username,
            password=password
        )
    elif HAS_IMPACKET:
        return _put_file_impacket(
            local_path, path, share, conn=conn, host=host, username=username,
            password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _delete_file_impacket(path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    conn.deleteFile(share, path)


def _delete_file_smbprotocol(path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    tree = conn.tree_connect(share)
    file_open = Open(tree, path)
    delete_msgs = [
        file_open.create(
            ImpersonationLevel.Impersonation,
            FilePipePrinterAccessMask.GENERIC_READ |
            FilePipePrinterAccessMask.DELETE,
            FileAttributes.FILE_ATTRIBUTE_NORMAL,
            ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE,
            CreateDisposition.FILE_OPEN,
            CreateOptions.FILE_NON_DIRECTORY_FILE |
            CreateOptions.FILE_DELETE_ON_CLOSE,
            send=False
        ),
        file_open.close(False, send=False)
    ]
    requests = conn.connection.send_compound([x[0] for x in delete_msgs],
                                        conn.session.session_id,
                                        tree.tree_connect_id, related=True)
    responses = []
    for i, request in enumerate(requests):
        # A SMBResponseException will be raised if something went wrong
        response = delete_msgs[i][1](request)
        responses.append(response)


def delete_file(path, share='C$', conn=None, host=None, username=None, password=None):
    if HAS_SMBPROTOCOL:
        return _delete_file_smbprotocol(path, share, conn=conn, host=host, username=username, password=password)
    elif HAS_IMPACKET:
        return _delete_file_impacket(path, share, conn=conn, host=host, username=username, password=password)
    raise MissingSmb("SMB library required (impacket or smbprotocol)")


def _delete_directory_impacket(path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    conn.deleteDirectory(share, path)


def _delete_directory_smbprotocol(path, share='C$', conn=None, host=None, username=None, password=None):
    if conn is None:
        conn = get_conn(host, username, password)
    if conn is False:
        return False
    log.debug("_delete_directory_smbprotocol - share: %s, path: %s", share, path)
    tree = conn.tree_connect(share)

    dir_open = Open(tree, path)
    delete_msgs = [
        dir_open.create(
            ImpersonationLevel.Impersonation,
            DirectoryAccessMask.DELETE,
            FileAttributes.FILE_ATTRIBUTE_DIRECTORY,
            0,
            CreateDisposition.FILE_OPEN,
            CreateOptions.FILE_DIRECTORY_FILE |
            CreateOptions.FILE_DELETE_ON_CLOSE,
            send=False
        ),
        dir_open.close(False, send=False)
    ]
    delete_reqs = conn.connection.send_compound([x[0] for x in delete_msgs],
                                           sid=conn.session.session_id,
                                           tid=tree.tree_connect_id,
                                           related=True)
    for i, request in enumerate(delete_reqs):
        # A SMBResponseException will be raised if something went wrong
        response = delete_msgs[i][1](request)


def delete_directory(path, share='C$', conn=None, host=None, username=None, password=None):
    if HAS_SMBPROTOCOL:
        return _delete_directory_smbprotocol(
            path, share, conn=conn, host=host, username=username, password=password
        )
    elif HAS_IMPACKET:
        return _delete_directory_impacket(
            path, share, conn=conn, host=host, username=username, password=password
        )
    raise MissingSmb("SMB library required (impacket or smbprotocol)")



# -*- coding: utf-8 -*-
'''
Generate the salt thin tarball from the installed python files
'''

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import copy
import logging
import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import zipfile

# Import third party libs
import jinja2
import yaml
import msgpack
import salt.ext.six as _six
import tornado

try:
    import zlib
except ImportError:
    zlib = None

# pylint: disable=import-error,no-name-in-module
try:
    import certifi
except ImportError:
    certifi = None

try:
    import singledispatch
except ImportError:
    singledispatch = None

try:
    import singledispatch_helpers
except ImportError:
    singledispatch_helpers = None

try:
    import backports_abc
except ImportError:
    import salt.ext.backports_abc as backports_abc

try:
    # New Jinja only
    import markupsafe
except ImportError:
    markupsafe = None

# pylint: enable=import-error,no-name-in-module

try:
    # Older python where the backport from pypi is installed
    from backports import ssl_match_hostname
except ImportError:
    # Other older python we use our bundled copy
    try:
        from salt.ext import ssl_match_hostname
    except ImportError:
        ssl_match_hostname = None

# Import salt libs
import salt
import salt.utils.files
import salt.utils.hashutils
import salt.utils.json
import salt.utils.path
import salt.utils.stringutils
import salt.exceptions
import salt.version

if _six.PY2:
    import concurrent
else:
    concurrent = None


log = logging.getLogger(__name__)


def _get_salt_call(*dirs, **namespaces):
    '''
    Return salt-call source, based on configuration.
    This will include additional namespaces for another versions of Salt,
    if needed (e.g. older interpreters etc).

    :dirs: List of directories to include in the system path
    :namespaces: Dictionary of namespace
    :return:
    '''
    template = '''# -*- coding: utf-8 -*-
import os
import sys

# Namespaces is a map: {namespace: major/minor version}, like {'2016.11.4': [2, 6]}
# Appears only when configured in Master configuration.
namespaces = %namespaces%

# Default system paths alongside the namespaces
syspaths = %dirs%
syspaths.append('py{0}'.format(sys.version_info[0]))

curr_ver = (sys.version_info[0], sys.version_info[1],)

namespace = ''
for ns in namespaces:
    if curr_ver == tuple(namespaces[ns]):
        namespace = ns
        break

for base in syspaths:
    sys.path.insert(0, os.path.join(os.path.dirname(__file__),
                                    namespace and os.path.join(namespace, base) or base))

if __name__ == '__main__':
    from salt.scripts import salt_call
    salt_call()
'''

    for tgt, cnt in [('%dirs%', dirs), ('%namespaces%', namespaces)]:
        template = template.replace(tgt, salt.utils.json.dumps(cnt))

    return salt.utils.stringutils.to_bytes(template)


def thin_path(cachedir):
    '''
    Return the path to the thin tarball
    '''
    return os.path.join(cachedir, 'thin', 'thin.tgz')


def _is_shareable(mod):
    '''
    Return True if module is share-able between major Python versions.

    :param mod:
    :return:
    '''
    # This list is subject to change
    shareable = ['salt', 'jinja2',
                 'msgpack', 'certifi']

    return os.path.basename(mod) in shareable


def _add_dependency(container, obj):
    '''
    Add a dependency to the top list.

    :param obj:
    :param is_file:
    :return:
    '''
    if os.path.basename(obj.__file__).split('.')[0] == '__init__':
        container.append(os.path.dirname(obj.__file__))
    else:
        container.append(obj.__file__.replace('.pyc', '.py'))


def gte():
    '''
    This function is called externally from the alternative
    Python interpreter from within _get_tops function.

    :param extra_mods:
    :param so_mods:
    :return:
    '''
    extra = salt.utils.json.loads(sys.argv[1])
    tops = get_tops(**extra)

    return salt.utils.json.dumps(tops, ensure_ascii=False)


def get_ext_tops(config):
    '''
    Get top directories for the dependencies, based on external configuration.

    :return:
    '''
    config = copy.deepcopy(config)
    alternatives = {}
    required = ['jinja2', 'yaml', 'tornado', 'msgpack']
    tops = []
    for ns, cfg in salt.ext.six.iteritems(config or {}):
        alternatives[ns] = cfg
        locked_py_version = cfg.get('py-version')
        err_msg = None
        if not locked_py_version:
            err_msg = 'Alternative Salt library: missing specific locked Python version'
        elif not isinstance(locked_py_version, (tuple, list)):
            err_msg = ('Alternative Salt library: specific locked Python version '
                       'should be a list of major/minor version')
        if err_msg:
            raise salt.exceptions.SaltSystemExit(err_msg)

        if cfg.get('dependencies') == 'inherit':
            # TODO: implement inheritance of the modules from _here_
            raise NotImplementedError('This feature is not yet implemented')
        else:
            for dep in cfg.get('dependencies'):
                mod = cfg['dependencies'][dep] or ''
                if not mod:
                    log.warning('Module %s has missing configuration', dep)
                    continue
                elif mod.endswith('.py') and not os.path.isfile(mod):
                    log.warning('Module %s configured with not a file or does not exist: %s', dep, mod)
                    continue
                elif not mod.endswith('.py') and not os.path.isfile(os.path.join(mod, '__init__.py')):
                    log.warning('Module %s is not a Python importable module with %s', dep, mod)
                    continue
                tops.append(mod)

                if dep in required:
                    required.pop(required.index(dep))

            required = ', '.join(required)
            if required:
                msg = 'Missing dependencies for the alternative version' \
                      ' in the external configuration: {}'.format(required)
                log.error(msg)
                raise salt.exceptions.SaltSystemExit(msg)
        alternatives[ns]['dependencies'] = tops
    return alternatives


def _get_ext_namespaces(config):
    '''
    Get namespaces from the existing configuration.

    :param config:
    :return:
    '''
    namespaces = {}
    if not config:
        return namespaces

    for ns in config:
        constraint_version = tuple(config[ns].get('py-version', []))
        if not constraint_version:
            raise salt.exceptions.SaltSystemExit("An alternative version is configured, but not defined "
                                                 "to what Python's major/minor version it should be constrained.")
        else:
            namespaces[ns] = constraint_version

    return namespaces


def get_tops(extra_mods='', so_mods=''):
    '''
    Get top directories for the dependencies, based on Python interpreter.

    :param extra_mods:
    :param so_mods:
    :return:
    '''
    tops = []
    for mod in [salt, jinja2, yaml, tornado, msgpack, certifi, singledispatch, concurrent,
                singledispatch_helpers, ssl_match_hostname, markupsafe, backports_abc]:
        if mod:
            log.debug('Adding module to the tops: "%s"', mod.__name__)
            _add_dependency(tops, mod)

    for mod in [m for m in extra_mods.split(',') if m]:
        if mod not in locals() and mod not in globals():
            try:
                locals()[mod] = __import__(mod)
                moddir, modname = os.path.split(locals()[mod].__file__)
                base, ext = os.path.splitext(modname)
                if base == '__init__':
                    tops.append(moddir)
                else:
                    tops.append(os.path.join(moddir, base + '.py'))
            except ImportError as err:
                log.exception(err)
                log.error('Unable to import extra-module "%s"', mod)

    for mod in [m for m in so_mods.split(',') if m]:
        try:
            locals()[mod] = __import__(mod)
            tops.append(locals()[mod].__file__)
        except ImportError as err:
            log.exception(err)
            log.error('Unable to import so-module "%s"', mod)

    return tops


def _get_supported_py_config(tops, extended_cfg):
    '''
    Based on the Salt SSH configuration, create a YAML configuration
    for the supported Python interpreter versions. This is then written into the thin.tgz
    archive and then verified by salt.client.ssh.ssh_py_shim.get_executable()

    Note: Minimum default of 2.x versions is 2.7 and 3.x is 3.0, unless specified in namespaces.

    :return:
    '''
    pymap = []
    for py_ver, tops in _six.iteritems(copy.deepcopy(tops)):
        py_ver = int(py_ver)
        if py_ver == 2:
            pymap.append('py2:2:7')
        elif py_ver == 3:
            pymap.append('py3:3:0')

    for ns, cfg in _six.iteritems(copy.deepcopy(extended_cfg) or {}):
        pymap.append('{}:{}:{}'.format(ns, *cfg.get('py-version')))
    pymap.append('')

    return salt.utils.stringutils.to_bytes(os.linesep.join(pymap))


def gen_thin(cachedir, extra_mods='', overwrite=False, so_mods='',
             python2_bin='python2', python3_bin='python3', absonly=True,
             compress='gzip', extended_cfg=None):
    '''
    Generate the salt-thin tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run thin.generate
        salt-run thin.generate mako
        salt-run thin.generate mako,wempy 1
        salt-run thin.generate overwrite=1
    '''
    if sys.version_info < (2, 6):
        raise salt.exceptions.SaltSystemExit('The minimum required python version to run salt-ssh is "2.6".')
    if compress not in ['gzip', 'zip']:
        log.warning('Unknown compression type: "%s". Falling back to "gzip" compression.', compress)
        compress = 'gzip'

    thindir = os.path.join(cachedir, 'thin')
    if not os.path.isdir(thindir):
        os.makedirs(thindir)
    thintar = os.path.join(thindir, 'thin.' + (compress == 'gzip' and 'tgz' or 'zip'))
    thinver = os.path.join(thindir, 'version')
    pythinver = os.path.join(thindir, '.thin-gen-py-version')
    salt_call = os.path.join(thindir, 'salt-call')
    pymap_cfg = os.path.join(thindir, 'supported-versions')
    code_checksum = os.path.join(thindir, 'code-checksum')
    digest_collector = salt.utils.hashutils.DigestCollector()

    with salt.utils.files.fopen(salt_call, 'wb') as fp_:
        fp_.write(_get_salt_call('pyall', **_get_ext_namespaces(extended_cfg)))

    if os.path.isfile(thintar):
        if not overwrite:
            if os.path.isfile(thinver):
                with salt.utils.files.fopen(thinver) as fh_:
                    overwrite = fh_.read() != salt.version.__version__
                if overwrite is False and os.path.isfile(pythinver):
                    with salt.utils.files.fopen(pythinver) as fh_:
                        overwrite = fh_.read() != str(sys.version_info[0])  # future lint: disable=blacklisted-function
            else:
                overwrite = True

        if overwrite:
            try:
                log.debug('Removing %s archive file', thintar)
                os.remove(thintar)
            except OSError as exc:
                log.error('Error while removing %s file: %s', thintar, exc)
                if os.path.exists(thintar):
                    raise salt.exceptions.SaltSystemExit('Unable to remove %s file. See logs for details.', thintar)
        else:
            return thintar
    if _six.PY3:
        # Let's check for the minimum python 2 version requirement, 2.6
        py_shell_cmd = "{} -c 'import sys;sys.stdout.write(\"%s.%s\\n\" % sys.version_info[:2]);'".format(python2_bin)
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, shell=True)
        stdout, _ = cmd.communicate()
        if cmd.returncode == 0:
            py2_version = tuple(int(n) for n in stdout.decode('utf-8').strip().split('.'))
            if py2_version < (2, 6):
                raise salt.exceptions.SaltSystemExit(
                    'The minimum required python version to run salt-ssh is "2.6".'
                    'The version reported by "{0}" is "{1}". Please try "salt-ssh '
                    '--python2-bin=<path-to-python-2.6-binary-or-higher>".'.format(python2_bin, stdout.strip()))
        else:
            log.error('Unable to detect Python-2 version')
            log.debug(stdout)

    tops_failure_msg = 'Failed %s tops for Python binary %s.'
    tops_py_version_mapping = {}
    tops = get_tops(extra_mods=extra_mods, so_mods=so_mods)
    tops_py_version_mapping[sys.version_info.major] = tops

    # Collect tops, alternative to 2.x version
    if _six.PY2 and sys.version_info.major == 2:
        # Get python 3 tops
        py_shell_cmd = "{0} -c 'import salt.utils.thin as t;print(t.gte())' '{1}'".format(
            python3_bin, salt.utils.json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = salt.utils.json.loads(stdout)
                tops_py_version_mapping['3'] = tops
            except ValueError as err:
                log.error(tops_failure_msg, 'parsing', python3_bin)
                log.exception(err)
        else:
            log.error(tops_failure_msg, 'collecting', python3_bin)
            log.debug(stderr)

    # Collect tops, alternative to 3.x version
    if _six.PY3 and sys.version_info.major == 3:
        # Get python 2 tops
        py_shell_cmd = "{0} -c 'import salt.utils.thin as t;print(t.gte())' '{1}'".format(
            python2_bin, salt.utils.json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = salt.utils.json.loads(stdout.decode('utf-8'))
                tops_py_version_mapping['2'] = tops
            except ValueError as err:
                log.error(tops_failure_msg, 'parsing', python2_bin)
                log.exception(err)
        else:
            log.error(tops_failure_msg, 'collecting', python2_bin)
            log.debug(stderr)

    with salt.utils.files.fopen(pymap_cfg, 'wb') as fp_:
        fp_.write(_get_supported_py_config(tops=tops_py_version_mapping, extended_cfg=extended_cfg))

    if compress == 'gzip':
        tfp = tarfile.open(thintar, 'w:gz', dereference=True)
    elif compress == 'zip':
        tfp = zipfile.ZipFile(thintar, 'w', compression=zlib and zipfile.ZIP_DEFLATED or zipfile.ZIP_STORED)
        tfp.add = tfp.write

    try:  # cwd may not exist if it was removed but salt was run from it
        start_dir = os.getcwd()
    except OSError:
        start_dir = None
    tempdir = None

    # Pack default data
    log.debug('Packing default libraries based on current Salt version')
    for py_ver, tops in _six.iteritems(tops_py_version_mapping):
        for top in tops:
            if absonly and not os.path.isabs(top):
                continue
            base = os.path.basename(top)
            top_dirname = os.path.dirname(top)
            if os.path.isdir(top_dirname):
                os.chdir(top_dirname)
            else:
                # This is likely a compressed python .egg
                tempdir = tempfile.mkdtemp()
                egg = zipfile.ZipFile(top_dirname)
                egg.extractall(tempdir)
                top = os.path.join(tempdir, base)
                os.chdir(tempdir)

            site_pkg_dir = _is_shareable(base) and 'pyall' or 'py{}'.format(py_ver)

            log.debug('Packing "%s" to "%s" destination', base, site_pkg_dir)
            if not os.path.isdir(top):
                # top is a single file module
                if os.path.exists(os.path.join(top_dirname, base)):
                    tfp.add(base, arcname=os.path.join(site_pkg_dir, base))
                continue
            for root, dirs, files in salt.utils.path.os_walk(base, followlinks=True):
                for name in files:
                    if not name.endswith(('.pyc', '.pyo')):
                        digest_collector.add(os.path.join(root, name))
                        arcname = os.path.join(site_pkg_dir, root, name)
                        if hasattr(tfp, 'getinfo'):
                            try:
                                # This is a little slow but there's no clear way to detect duplicates
                                tfp.getinfo(os.path.join(site_pkg_dir, root, name))
                                arcname = None
                            except KeyError:
                                log.debug('ZIP: Unable to add "%s" with "getinfo"', arcname)
                        if arcname:
                            tfp.add(os.path.join(root, name), arcname=arcname)

            if tempdir is not None:
                shutil.rmtree(tempdir)
                tempdir = None

    # Pack alternative data
    if extended_cfg:
        log.debug('Packing libraries based on alternative Salt versions')
    for ns, cfg in _six.iteritems(get_ext_tops(extended_cfg)):
        tops = [cfg.get('path')] + cfg.get('dependencies')
        py_ver_major, py_ver_minor = cfg.get('py-version')
        for top in tops:
            base, top_dirname = os.path.basename(top), os.path.dirname(top)
            os.chdir(top_dirname)
            site_pkg_dir = _is_shareable(base) and 'pyall' or 'py{0}'.format(py_ver_major)
            log.debug('Packing alternative "%s" to "%s/%s" destination', base, ns, site_pkg_dir)
            if not os.path.isdir(top):
                # top is a single file module
                if os.path.exists(os.path.join(top_dirname, base)):
                    tfp.add(base, arcname=os.path.join(ns, site_pkg_dir, base))
                continue
            for root, dirs, files in salt.utils.path.os_walk(base, followlinks=True):
                for name in files:
                    if not name.endswith(('.pyc', '.pyo')):
                        digest_collector.add(os.path.join(root, name))
                        arcname = os.path.join(ns, site_pkg_dir, root, name)
                        if hasattr(tfp, 'getinfo'):
                            try:
                                tfp.getinfo(os.path.join(site_pkg_dir, root, name))
                                arcname = None
                            except KeyError:
                                log.debug('ZIP: Unable to add "%s" with "getinfo"', arcname)
                        if arcname:
                            tfp.add(os.path.join(root, name), arcname=arcname)

    os.chdir(thindir)
    with salt.utils.files.fopen(thinver, 'w+') as fp_:
        fp_.write(salt.version.__version__)
    with salt.utils.files.fopen(pythinver, 'w+') as fp_:
        fp_.write(str(sys.version_info.major))  # future lint: disable=blacklisted-function
    with salt.utils.files.fopen(code_checksum, 'w+') as fp_:
        fp_.write(digest_collector.digest())
    os.chdir(os.path.dirname(thinver))

    for fname in ['version', '.thin-gen-py-version', 'salt-call', 'supported-versions', 'code-checksum']:
        tfp.add(fname)

    if start_dir:
        os.chdir(start_dir)
    tfp.close()

    return thintar


def thin_sum(cachedir, form='sha1'):
    '''
    Return the checksum of the current thin tarball
    '''
    thintar = gen_thin(cachedir)
    code_checksum_path = os.path.join(cachedir, 'thin', 'code-checksum')
    if os.path.isfile(code_checksum_path):
        with salt.utils.files.fopen(code_checksum_path, 'r') as fh:
            code_checksum = "'{0}'".format(fh.read().strip())
    else:
        code_checksum = "'0'"

    return code_checksum, salt.utils.hashutils.get_hash(thintar, form)


def gen_min(cachedir, extra_mods='', overwrite=False, so_mods='',
            python2_bin='python2', python3_bin='python3'):
    '''
    Generate the salt-min tarball and print the location of the tarball
    Optional additional mods to include (e.g. mako) can be supplied as a comma
    delimited string.  Permits forcing an overwrite of the output file as well.

    CLI Example:

    .. code-block:: bash

        salt-run min.generate
        salt-run min.generate mako
        salt-run min.generate mako,wempy 1
        salt-run min.generate overwrite=1
    '''
    mindir = os.path.join(cachedir, 'min')
    if not os.path.isdir(mindir):
        os.makedirs(mindir)
    mintar = os.path.join(mindir, 'min.tgz')
    minver = os.path.join(mindir, 'version')
    pyminver = os.path.join(mindir, '.min-gen-py-version')
    salt_call = os.path.join(mindir, 'salt-call')
    with salt.utils.files.fopen(salt_call, 'wb') as fp_:
        fp_.write(_get_salt_call())
    if os.path.isfile(mintar):
        if not overwrite:
            if os.path.isfile(minver):
                with salt.utils.files.fopen(minver) as fh_:
                    overwrite = fh_.read() != salt.version.__version__
                if overwrite is False and os.path.isfile(pyminver):
                    with salt.utils.files.fopen(pyminver) as fh_:
                        overwrite = fh_.read() != str(sys.version_info[0])  # future lint: disable=blacklisted-function
            else:
                overwrite = True

        if overwrite:
            try:
                os.remove(mintar)
            except OSError:
                pass
        else:
            return mintar
    if _six.PY3:
        # Let's check for the minimum python 2 version requirement, 2.6
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; import sys; '
            'print("{0}.{1}".format(*(sys.version_info[:2])));\''
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, shell=True)
        stdout, _ = cmd.communicate()
        if cmd.returncode == 0:
            py2_version = tuple(int(n) for n in stdout.decode('utf-8').strip().split('.'))
            if py2_version < (2, 6):
                # Bail!
                raise salt.exceptions.SaltSystemExit(
                    'The minimum required python version to run salt-ssh is "2.6".'
                    'The version reported by "{0}" is "{1}". Please try "salt-ssh '
                    '--python2-bin=<path-to-python-2.6-binary-or-higher>".'.format(python2_bin,
                                                                                stdout.strip())
                )
    elif sys.version_info < (2, 6):
        # Bail! Though, how did we reached this far in the first place.
        raise salt.exceptions.SaltSystemExit(
            'The minimum required python version to run salt-ssh is "2.6".'
        )

    tops_py_version_mapping = {}
    tops = get_tops(extra_mods=extra_mods, so_mods=so_mods)
    if _six.PY2:
        tops_py_version_mapping['2'] = tops
    else:
        tops_py_version_mapping['3'] = tops

    # TODO: Consider putting known py2 and py3 compatible libs in it's own sharable directory.
    #       This would reduce the min size.
    if _six.PY2 and sys.version_info[0] == 2:
        # Get python 3 tops
        py_shell_cmd = (
            python3_bin + ' -c \'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))), ensure_ascii=False)); exit(0);\' '
            '\'{0}\''.format(salt.utils.json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = salt.utils.json.loads(stdout)
                tops_py_version_mapping['3'] = tops
            except ValueError:
                pass
    if _six.PY3 and sys.version_info[0] == 3:
        # Get python 2 tops
        py_shell_cmd = (
            python2_bin + ' -c \'from __future__ import print_function; '
            'import sys; import json; import salt.utils.thin; '
            'print(json.dumps(salt.utils.thin.get_tops(**(json.loads(sys.argv[1]))), ensure_ascii=False)); exit(0);\' '
            '\'{0}\''.format(salt.utils.json.dumps({'extra_mods': extra_mods, 'so_mods': so_mods}))
        )
        cmd = subprocess.Popen(py_shell_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = cmd.communicate()
        if cmd.returncode == 0:
            try:
                tops = salt.utils.json.loads(stdout.decode('utf-8'))
                tops_py_version_mapping['2'] = tops
            except ValueError:
                pass

    tfp = tarfile.open(mintar, 'w:gz', dereference=True)
    try:  # cwd may not exist if it was removed but salt was run from it
        start_dir = os.getcwd()
    except OSError:
        start_dir = None
    tempdir = None

    # This is the absolute minimum set of files required to run salt-call
    min_files = (
        'salt/__init__.py',
        'salt/utils',
        'salt/utils/__init__.py',
        'salt/utils/atomicfile.py',
        'salt/utils/validate',
        'salt/utils/validate/__init__.py',
        'salt/utils/validate/path.py',
        'salt/utils/decorators',
        'salt/utils/decorators/__init__.py',
        'salt/utils/cache.py',
        'salt/utils/xdg.py',
        'salt/utils/odict.py',
        'salt/utils/minions.py',
        'salt/utils/dicttrim.py',
        'salt/utils/sdb.py',
        'salt/utils/migrations.py',
        'salt/utils/files.py',
        'salt/utils/parsers.py',
        'salt/utils/locales.py',
        'salt/utils/lazy.py',
        'salt/utils/s3.py',
        'salt/utils/dictupdate.py',
        'salt/utils/verify.py',
        'salt/utils/args.py',
        'salt/utils/kinds.py',
        'salt/utils/xmlutil.py',
        'salt/utils/debug.py',
        'salt/utils/jid.py',
        'salt/utils/openstack',
        'salt/utils/openstack/__init__.py',
        'salt/utils/openstack/swift.py',
        'salt/utils/asynchronous.py',
        'salt/utils/process.py',
        'salt/utils/jinja.py',
        'salt/utils/rsax931.py',
        'salt/utils/context.py',
        'salt/utils/minion.py',
        'salt/utils/error.py',
        'salt/utils/aws.py',
        'salt/utils/timed_subprocess.py',
        'salt/utils/zeromq.py',
        'salt/utils/schedule.py',
        'salt/utils/url.py',
        'salt/utils/yamlencoding.py',
        'salt/utils/network.py',
        'salt/utils/http.py',
        'salt/utils/gzip_util.py',
        'salt/utils/vt.py',
        'salt/utils/templates.py',
        'salt/utils/aggregation.py',
        'salt/utils/yaml.py',
        'salt/utils/yamldumper.py',
        'salt/utils/yamlloader.py',
        'salt/utils/event.py',
        'salt/utils/state.py',
        'salt/serializers',
        'salt/serializers/__init__.py',
        'salt/serializers/yamlex.py',
        'salt/template.py',
        'salt/_compat.py',
        'salt/loader.py',
        'salt/client',
        'salt/client/__init__.py',
        'salt/ext',
        'salt/ext/__init__.py',
        'salt/ext/six.py',
        'salt/ext/ipaddress.py',
        'salt/version.py',
        'salt/syspaths.py',
        'salt/defaults',
        'salt/defaults/__init__.py',
        'salt/defaults/exitcodes.py',
        'salt/renderers',
        'salt/renderers/__init__.py',
        'salt/renderers/jinja.py',
        'salt/renderers/yaml.py',
        'salt/modules',
        'salt/modules/__init__.py',
        'salt/modules/test.py',
        'salt/modules/selinux.py',
        'salt/modules/cmdmod.py',
        'salt/modules/saltutil.py',
        'salt/minion.py',
        'salt/pillar',
        'salt/pillar/__init__.py',
        'salt/textformat.py',
        'salt/log',
        'salt/log/__init__.py',
        'salt/log/handlers',
        'salt/log/handlers/__init__.py',
        'salt/log/mixins.py',
        'salt/log/setup.py',
        'salt/cli',
        'salt/cli/__init__.py',
        'salt/cli/caller.py',
        'salt/cli/daemons.py',
        'salt/cli/salt.py',
        'salt/cli/call.py',
        'salt/fileserver',
        'salt/fileserver/__init__.py',
        'salt/transport',
        'salt/transport/__init__.py',
        'salt/transport/client.py',
        'salt/exceptions.py',
        'salt/grains',
        'salt/grains/__init__.py',
        'salt/grains/extra.py',
        'salt/scripts.py',
        'salt/state.py',
        'salt/fileclient.py',
        'salt/crypt.py',
        'salt/config.py',
        'salt/beacons',
        'salt/beacons/__init__.py',
        'salt/payload.py',
        'salt/output',
        'salt/output/__init__.py',
        'salt/output/nested.py',
    )

    for py_ver, tops in _six.iteritems(tops_py_version_mapping):
        for top in tops:
            base = os.path.basename(top)
            top_dirname = os.path.dirname(top)
            if os.path.isdir(top_dirname):
                os.chdir(top_dirname)
            else:
                # This is likely a compressed python .egg
                tempdir = tempfile.mkdtemp()
                egg = zipfile.ZipFile(top_dirname)
                egg.extractall(tempdir)
                top = os.path.join(tempdir, base)
                os.chdir(tempdir)
            if not os.path.isdir(top):
                # top is a single file module
                tfp.add(base, arcname=os.path.join('py{0}'.format(py_ver), base))
                continue
            for root, dirs, files in salt.utils.path.os_walk(base, followlinks=True):
                for name in files:
                    if name.endswith(('.pyc', '.pyo')):
                        continue
                    if root.startswith('salt') and os.path.join(root, name) not in min_files:
                        continue
                    tfp.add(os.path.join(root, name),
                            arcname=os.path.join('py{0}'.format(py_ver), root, name))
            if tempdir is not None:
                shutil.rmtree(tempdir)
                tempdir = None

    os.chdir(mindir)
    tfp.add('salt-call')
    with salt.utils.files.fopen(minver, 'w+') as fp_:
        fp_.write(salt.version.__version__)
    with salt.utils.files.fopen(pyminver, 'w+') as fp_:
        fp_.write(str(sys.version_info[0]))  # future lint: disable=blacklisted-function
    os.chdir(os.path.dirname(minver))
    tfp.add('version')
    tfp.add('.min-gen-py-version')
    if start_dir:
        os.chdir(start_dir)
    tfp.close()
    return mintar


def min_sum(cachedir, form='sha1'):
    '''
    Return the checksum of the current thin tarball
    '''
    mintar = gen_min(cachedir)
    return salt.utils.hashutils.get_hash(mintar, form)



# -*- coding: utf-8 -*-
'''
This is the default list matcher.
'''
from __future__ import absolute_import, print_function, unicode_literals
import logging

log = logging.getLogger(__name__)


def match(tgt):
    '''
    Determines if this host is on the list
    '''
    try:
        if ',' + __opts__['id'] + ',' in tgt \
                or tgt.startswith(__opts__['id'] + ',') \
                or tgt.endswith(',' + __opts__['id']):
            return True
        # tgt is a string, which we know because the if statement above did not
        # cause one of the exceptions being caught. Therefore, look for an
        # exact match. (e.g. salt -L foo test.ping)
        return __opts__['id'] == tgt
    except (AttributeError, TypeError):
        # tgt is not a string, maybe it's a sequence type?
        try:
            return __opts__['id'] in tgt
        except Exception:
            # tgt was likely some invalid type
            return False

    # We should never get here based on the return statements in the logic
    # above. If we do, it is because something above changed, and should be
    # considered as a bug. Log a warning to help us catch this.
    log.warning(
        'List matcher unexpectedly did not return, for target %s, '
        'this is probably a bug.', tgt
    )
    return False



# -*- coding: utf-8 -*-
'''
    :codeauthor: Mike Place <mp@saltstack.com>
'''

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import os
import errno
import socket
import logging

import tornado.gen
import tornado.ioloop
import tornado.testing

import salt.config
import salt.exceptions
import salt.transport.ipc
import salt.transport.server
import salt.transport.client
import salt.utils.platform

from salt.ext import six
from salt.ext.six.moves import range

# Import Salt Testing libs
from tests.support.mock import MagicMock
from tests.support.paths import TMP
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@skipIf(salt.utils.platform.is_windows(), 'Windows does not support Posix IPC')
class BaseIPCReqCase(tornado.testing.AsyncTestCase):
    '''
    Test the req server/client pair
    '''
    def setUp(self):
        super(BaseIPCReqCase, self).setUp()
        #self._start_handlers = dict(self.io_loop._handlers)
        self.socket_path = os.path.join(TMP, 'ipc_test.ipc')

        self.server_channel = salt.transport.ipc.IPCMessageServer(
            self.socket_path,
            io_loop=self.io_loop,
            payload_handler=self._handle_payload,
        )
        self.server_channel.start()

        self.payloads = []

    def tearDown(self):
        super(BaseIPCReqCase, self).tearDown()
        #failures = []
        try:
            self.server_channel.close()
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        os.unlink(self.socket_path)
        #for k, v in six.iteritems(self.io_loop._handlers):
        #    if self._start_handlers.get(k) != v:
        #        failures.append((k, v))
        #if len(failures) > 0:
        #    raise Exception('FDs still attached to the IOLoop: {0}'.format(failures))
        del self.payloads
        del self.socket_path
        del self.server_channel
        #del self._start_handlers

    @tornado.gen.coroutine
    def _handle_payload(self, payload, reply_func):
        self.payloads.append(payload)
        yield reply_func(payload)
        if isinstance(payload, dict) and payload.get('stop'):
            self.stop()


class IPCMessageClient(BaseIPCReqCase):
    '''
    Test all of the clear msg stuff
    '''

    def _get_channel(self):
        channel = salt.transport.ipc.IPCMessageClient(
            socket_path=self.socket_path,
            io_loop=self.io_loop,
        )
        channel.connect(callback=self.stop)
        self.wait()
        return channel

    def setUp(self):
        super(IPCMessageClient, self).setUp()
        self.channel = self._get_channel()

    def tearDown(self):
        super(IPCMessageClient, self).tearDown()
        try:
            # Make sure we close no matter what we've done in the tests
            del self.channel
        except socket.error as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise

    def test_singleton(self):
        channel = self._get_channel()
        assert self.channel is channel
        # Delete the local channel. Since there's still one more refefence
        # __del__ wasn't called
        del channel
        assert self.channel
        msg = {'foo': 'bar', 'stop': True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[0], msg)

    def test_last_singleton_instance_closes(self):
        channel = self._get_channel()
        msg = {'foo': 'bar', 'stop': True}
        log.debug('Sending msg1')
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[0], msg)
        channel.close()
        # Since this is a singleton, and only the last singleton instance
        # should actually close the connection, the next code should still
        # work and not timeout
        msg = {'bar': 'foo', 'stop': True}
        log.debug('Sending msg2')
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[1], msg)

    def test_basic_send(self):
        msg = {'foo': 'bar', 'stop': True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(self.payloads[0], msg)

    def test_many_send(self):
        msgs = []
        self.server_channel.stream_handler = MagicMock()

        for i in range(0, 1000):
            msgs.append('test_many_send_{0}'.format(i))

        for i in msgs:
            self.channel.send(i)
        self.channel.send({'stop': True})
        self.wait()
        self.assertEqual(self.payloads[:-1], msgs)

    def test_very_big_message(self):
        long_str = ''.join([six.text_type(num) for num in range(10**5)])
        msg = {'long_str': long_str, 'stop': True}
        self.channel.send(msg)
        self.wait()
        self.assertEqual(msg, self.payloads[0])

    def test_multistream_sends(self):
        local_channel = self._get_channel()

        for c in (self.channel, local_channel):
            c.send('foo')

        self.channel.send({'stop': True})
        self.wait()
        self.assertEqual(self.payloads[:-1], ['foo', 'foo'])

    def test_multistream_errors(self):
        local_channel = self._get_channel()

        for c in (self.channel, local_channel):
            c.send(None)

        for c in (self.channel, local_channel):
            c.send('foo')

        self.channel.send({'stop': True})
        self.wait()
        self.assertEqual(self.payloads[:-1], [None, None, 'foo', 'foo'])



# -*- coding: utf-8 -*-
'''
Test salt-call --proxyid commands

tests.integration.proxy.test_shell
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'''

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.utils.json as json

# Import salt tests libs
from tests.support.case import ShellCase


class ProxyCallerSimpleTestCase(ShellCase):
    '''
    Test salt-call --proxyid <proxyid> commands
    '''
    @staticmethod
    def _load_return(ret):
        return json.loads('\n'.join(ret))

    def test_can_it_ping(self):
        '''
        Ensure the proxy can ping
        '''
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json test.ping'))
        self.assertEqual(ret['local'], True)

    def test_list_pkgs(self):
        '''
        Package test 1, really just tests that the virtual function capability
        is working OK.
        '''
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json pkg.list_pkgs'))
        self.assertIn('coreutils', ret['local'])
        self.assertIn('apache', ret['local'])
        self.assertIn('redbull', ret['local'])

    def test_upgrade(self):
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json pkg.upgrade'))
        self.assertEqual(ret['local']['coreutils']['new'], '2.0')
        self.assertEqual(ret['local']['redbull']['new'], '1000.99')

    def test_service_list(self):
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json service.list'))
        self.assertIn('ntp', ret['local'])

    def test_service_start(self):
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json service.start samba'))
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json service.status samba'))
        self.assertTrue(ret)

    def test_service_get_all(self):
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json service.get_all'))
        self.assertIn('samba', ret['local'])

    def test_grains_items(self):
        ret = self._load_return(self.run_call('--proxyid proxytest --out=json grains.items'))
        self.assertEqual(ret['local']['kernel'], 'proxy')
        self.assertEqual(ret['local']['kernelrelease'], 'proxy')



# -*- coding: utf-8 -*-
'''
Integration tests for the lxd states
'''
# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Lxd Test Case
import tests.integration.states.test_lxd


class LxdProfileTestCase(tests.integration.states.test_lxd.LxdTestCase):

    def tearDown(self):
        self.run_state(
            'lxd_profile.absent',
            name='test-profile',
        )

    def test_02__create_profile(self):
        self.run_state(
            'lxd_profile.absent',
            name='test-profile',
        )
        ret = self.run_state(
            'lxd_profile.present',
            name='test-profile',
            config=[
                {'key': 'boot.autostart', 'value': 1},
            ],
        )
        name = 'lxd_profile_|-test-profile_|-test-profile_|-present'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes'] == {'created': 'Profile "test-profile" has been created'}

    def test_03__change_profile(self):
        self.run_state(
            'lxd_profile.present',
            name='test-profile',
            config=[
                {'key': 'boot.autostart', 'value': 1},
            ],
        )
        ret = self.run_state(
            'lxd_profile.present',
            name='test-profile',
            config=[
                {'key': 'boot.autostart', 'value': 1},
                {'key': 'security.privileged', 'value': '1'},
            ],
        )
        name = 'lxd_profile_|-test-profile_|-test-profile_|-present'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes']['config'] == {
            'security.privileged': 'Added config key "security.privileged" = "1"'
        }

    def test_04__delete_profile(self):
        self.run_state(
            'lxd_profile.present',
            name='test-profile',
            config=[
                {'key': 'boot.autostart', 'value': 1},
            ],
        )
        ret = self.run_state(
            'lxd_profile.absent',
            name='test-profile',
        )
        name = 'lxd_profile_|-test-profile_|-test-profile_|-absent'
        self.assertSaltTrueReturn(ret)
        assert name in ret
        assert ret[name]['changes'] == {'removed': 'Profile "test-profile" has been deleted.'}



# -*- coding: utf-8 -*-
'''
Test utility methods that communicate with SMB shares.
'''
from __future__ import absolute_import
import getpass
import logging
import os
import signal
import subprocess
import tempfile
import time

import salt.utils.files
import salt.utils.path
import salt.utils.smb

from tests.support.unit import skipIf
from tests.support.case import TestCase

log = logging.getLogger(__name__)
CONFIG = (
    '[global]\n'
    'realm = saltstack.com\n'
    'interfaces = lo 127.0.0.0/8\n'
    'smb ports = 1445\n'
    'log level = 2\n'
    'map to guest = Bad User\n'
    'enable core files = no\n'
    'passdb backend = smbpasswd\n'
    'smb passwd file = {passwdb}\n'
    'lock directory = {samba_dir}\n'
    'state directory = {samba_dir}\n'
    'cache directory = {samba_dir}\n'
    'pid directory = {samba_dir}\n'
    'private dir = {samba_dir}\n'
    'ncalrpc dir = {samba_dir}\n'
    'socket options = IPTOS_LOWDELAY TCP_NODELAY\n'
    'min receivefile size = 0\n'
    'write cache size = 0\n'
    'client ntlmv2 auth = no\n'
    'client min protocol = SMB3_11\n'
    'client plaintext auth = no\n'
    '\n'
    '[public]\n'
    'path = {public_dir}\n'
    'read only = no\n'
    'guest ok = no\n'
    'writeable = yes\n'
    'force user = {user}\n'
)
TBE = (
    '{}:0:XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX:AC8E657F8'
    '3DF82BEEA5D43BDAF7800CC:[U          ]:LCT-507C14C7:'
)


def which_smbd():
    '''
    Find the smbd executable and cache the result if it exits.
    '''
    if hasattr(which_smbd, 'cached_result'):
        return which_smbd.cached_result
    smbd = salt.utils.path.which('smbd')
    if smbd:
        which_smbd.cached_result = smbd
    return smbd


@skipIf(not which_smbd(), 'Skip when no smbd binary found')
class TestSmb(TestCase):

    _smbd = None

    @staticmethod
    def check_pid(pid):
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    @classmethod
    def setUpClass(cls):
        tmpdir = tempfile.mkdtemp()
        cls.samba_dir = os.path.join(tmpdir, 'samba')
        cls.public_dir = os.path.join(tmpdir, 'public')
        os.makedirs(cls.samba_dir)
        os.makedirs(cls.public_dir)
        os.chmod(cls.samba_dir, 0o775)
        os.chmod(cls.public_dir, 0o775)
        passwdb = os.path.join(tmpdir, 'passwdb')
        cls.username = getpass.getuser()
        with salt.utils.files.fopen(passwdb, 'w') as fp:
            fp.write(TBE.format(cls.username))
        samba_conf = os.path.join(tmpdir, 'smb.conf')
        with salt.utils.files.fopen(samba_conf, 'w') as fp:
            fp.write(
                CONFIG.format(
                    samba_dir=cls.samba_dir,
                    public_dir=cls.public_dir,
                    passwdb=passwdb,
                    user=cls.username,
                )
            )
        cls._smbd = subprocess.Popen(
            '{0} -FS -P0 -s {1}'.format(which_smbd(), samba_conf),
            shell=True
        )
        time.sleep(1)
        pidfile = os.path.join(cls.samba_dir, 'smbd.pid')
        with salt.utils.files.fopen(pidfile, 'r') as fp:
            cls._pid = int(fp.read().strip())
        if not cls.check_pid(cls._pid):
            raise Exception('Unable to locate smbd\'s pid file')

    @classmethod
    def tearDownClass(cls):
        log.warning('teardown')
        os.kill(cls._pid, signal.SIGTERM)

    def test_write_file(self):
        '''
        Transfer a file over SMB
        '''
        name = 'test_write_file.txt'
        content = 'write test file content'
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)

        local_path = tempfile.mktemp()
        with salt.utils.files.fopen(local_path, 'w') as fp:
            fp.write(content)
        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.put_file(local_path, name, 'public', conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, 'r') as fp:
            result = fp.read()
        assert result == content

    def test_write_str(self):
        '''
        Write a string to a file over SMB
        '''
        name = 'test_write_str.txt'
        content = 'write test file content'
        share_path = os.path.join(self.public_dir, name)
        assert not os.path.exists(share_path)
        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.put_str(content, name, 'public', conn=conn)
        conn.close()

        assert os.path.exists(share_path)
        with salt.utils.files.fopen(share_path, 'r') as fp:
            result = fp.read()
        assert result == content

    def test_delete_file(self):
        '''
        Validate deletion of files over SMB
        '''
        name = 'test_delete_file.txt'
        content = 'read test file content'
        share_path = os.path.join(self.public_dir, name)
        with salt.utils.files.fopen(share_path, 'w') as fp:
            fp.write(content)
        assert os.path.exists(share_path)

        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.delete_file(name, 'public', conn=conn)
        conn.close()

        assert not os.path.exists(share_path)

    def test_mkdirs(self):
        '''
        Create directories over SMB
        '''
        dir_name = 'mkdirs/test'
        share_path = os.path.join(self.public_dir, dir_name)
        assert not os.path.exists(share_path)

        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.mkdirs(dir_name, 'public', conn=conn)
        conn.close()

        assert os.path.exists(share_path)

    def test_delete_dirs(self):
        '''
        Validate deletion of directoreies over SMB
        '''
        dir_name = 'deldirs'
        subdir_name = 'deldirs/test'
        local_path = os.path.join(self.public_dir, subdir_name)
        os.makedirs(local_path)
        assert os.path.exists(local_path)

        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.delete_directory(subdir_name, 'public', conn=conn)
        conn.close()

        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        salt.utils.smb.delete_directory(dir_name, 'public', conn=conn)
        conn.close()

        assert not os.path.exists(local_path)
        assert not os.path.exists(os.path.join(self.public_dir, dir_name))

    def test_connection(self):
        '''
        Validate creation of an SMB connection
        '''
        conn = salt.utils.smb.get_conn('127.0.0.1', self.username, 'foo', port=1445)
        conn.close()



# -*- coding: utf-8 -*-
'''
This is the default glob matcher function.
'''
from __future__ import absolute_import, print_function, unicode_literals

import fnmatch
from salt.ext import six  # pylint: disable=3rd-party-module-not-gated


def match(tgt):
    '''
    Returns true if the passed glob matches the id
    '''
    if not isinstance(tgt, six.string_types):
        return False

    return fnmatch.fnmatch(__opts__['id'], tgt)



# -*- coding: utf-8 -*-
'''
unittests for terraform roster
'''
# Import Python libs
from __future__ import absolute_import, unicode_literals
import os.path

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase, skipIf
from tests.support.runtests import RUNTIME_VARS
from tests.support.mock import (
    patch,
    NO_MOCK,
    NO_MOCK_REASON
)

# Import Salt Libs
import salt.config
import salt.loader
from salt.roster import terraform


@skipIf(NO_MOCK, NO_MOCK_REASON)
class TerraformTestCase(TestCase, LoaderModuleMockMixin):
    '''
    Test cases for salt.roster.terraform
    '''
    def setup_loader_modules(self):
        opts = salt.config.master_config(os.path.join(RUNTIME_VARS.TMP_CONF_DIR, 'master'))
        utils = salt.loader.utils(opts, whitelist=['roster_matcher'])
        return {
            terraform: {
                '__utils__': utils,
                '__opts__': {},
            }
        }

    def test_default_output(self):
        '''
        Test the output of a fixture tfstate file wich contains libvirt
        resources.
        '''
        tfstate = os.path.join(os.path.dirname(__file__), 'terraform.data', 'terraform.tfstate')
        pki_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'terraform.data'))

        with patch.dict(terraform.__opts__, {'roster_file': tfstate, 'pki_dir': pki_dir}):
            expected_result = {
                'db0': {
                    'host': '192.168.122.174',
                    'user': 'root',
                    'passwd': 'dbpw',
                    'tty': True,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')},
                'db1': {
                    'host': '192.168.122.190',
                    'user': 'root',
                    'passwd': 'dbpw',
                    'tty': True,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')},
                'web0': {
                    'host': '192.168.122.106',
                    'user': 'root',
                    'passwd': 'linux',
                    'timeout': 22,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')},
                'web1': {
                    'host': '192.168.122.235',
                    'user': 'root',
                    'passwd': 'linux',
                    'timeout': 22,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')}
            }

            ret = terraform.targets('*')
            self.assertDictEqual(expected_result, ret)

    def test_default_matching(self):
        '''
        Test the output of a fixture tfstate file wich contains libvirt
        resources using matching
        '''
        tfstate = os.path.join(os.path.dirname(__file__), 'terraform.data', 'terraform.tfstate')
        pki_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'terraform.data'))

        with patch.dict(terraform.__opts__, {'roster_file': tfstate, 'pki_dir': pki_dir}):
            expected_result = {
                'web0': {
                    'host': '192.168.122.106',
                    'user': 'root',
                    'passwd': 'linux',
                    'timeout': 22,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')},
                'web1': {
                    'host': '192.168.122.235',
                    'user': 'root',
                    'passwd': 'linux',
                    'timeout': 22,
                    'priv': os.path.join(pki_dir, 'ssh', 'salt-ssh.rsa')}
            }

            ret = terraform.targets('*web*')
            self.assertDictEqual(expected_result, ret)



# -*- coding: utf-8 -*-

# Import Python libs
from __future__ import absolute_import

# Import Salt Libs
import salt.modules.vmctl as vmctl

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase
from tests.support.mock import (
    MagicMock,
    patch,
)


class VmctlTestCase(TestCase, LoaderModuleMockMixin):
    '''
    test modules.vmctl functions
    '''
    def setup_loader_modules(self):
        return {vmctl: {}}

    def test_create_disk(self):
        '''
        Tests creating a new disk image.
        '''
        ret = {}
        ret['stdout'] = 'vmctl: imagefile created'
        ret['stderr'] = ''
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(vmctl.create_disk('/path/to/disk.img', '1G'))

    def test_load(self):
        '''
        Tests loading a configuration file.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(vmctl.load('/etc/vm.switches.conf'))

    def test_reload(self):
        '''
        Tests reloading the configuration.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertTrue(vmctl.reload())

    def test_reset(self):
        '''
        Tests resetting VMM.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.reset()
            mock_cmd.assert_called_once_with(['vmctl', 'reset'],
                output_loglevel='trace', python_shell=False)
            self.assertTrue(res)

    def test_reset_vms(self):
        '''
        Tests resetting VMs.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.reset(vms=True)
            mock_cmd.assert_called_once_with(['vmctl', 'reset', 'vms'],
                output_loglevel='trace', python_shell=False)
            self.assertTrue(res)

    def test_reset_switches(self):
        '''
        Tests resetting switches.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.reset(switches=True)
            mock_cmd.assert_called_once_with(['vmctl', 'reset', 'switches'],
                output_loglevel='trace', python_shell=False)
            self.assertTrue(res)

    def test_reset_all(self):
        '''
        Tests resetting all.
        '''
        ret = {}
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.reset(all=True)
            mock_cmd.assert_called_once_with(['vmctl', 'reset', 'all'],
                output_loglevel='trace', python_shell=False)
            self.assertTrue(res)

    def test_start_existing_vm(self):
        '''
        Tests starting a VM that is already defined.
        '''
        ret = {}
        ret['stderr'] = 'vmctl: started vm 4 successfully, tty /dev/ttyp4'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {'changes': True, 'console': '/dev/ttyp4'}
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertDictEqual(expected, vmctl.start('4'))

    def test_start_new_vm(self):
        '''
        Tests starting a new VM.
        '''
        ret = {}
        ret['stderr'] = 'vmctl: started vm 4 successfully, tty /dev/ttyp4'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        mock_status = MagicMock(return_value={})
        expected = {'changes': True, 'console': '/dev/ttyp4'}
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            with patch('salt.modules.vmctl.status', mock_status):
                res = vmctl.start('web1', bootpath='/bsd.rd', nics=2, disk='/disk.img')
                mock_cmd.assert_called_once_with(['vmctl', 'start', 'web1', '-i 2', '-b', '/bsd.rd', '-d', '/disk.img'],
                    output_loglevel='trace', python_shell=False)
                self.assertDictEqual(expected, res)

    def test_status(self):
        '''
        Tests getting status for all VMs.
        '''
        ret = {}
        ret['stdout'] = '   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n' \
                        '    1   123     1    2.9G     150M  ttyp5       john web1 - stopping\n' \
                        '    2   456     1    512M     301M  ttyp4       paul web2\n' \
                        '    3     -     1    512M       -       -       george web3\n'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {
            'web1': {
                'curmem': '150M',
                'id': '1',
                'maxmem': '2.9G',
                'owner': 'john',
                'pid': '123',
                'state': 'stopping',
                'tty': 'ttyp5',
                'vcpus': '1'
            },
            'web2': {
                'curmem': '301M',
                'id': '2',
                'maxmem': '512M',
                'owner': 'paul',
                'pid': '456',
                'state': 'running',
                'tty': 'ttyp4',
                'vcpus': '1'
            },
            'web3': {
                'curmem': '-',
                'id': '3',
                'maxmem': '512M',
                'owner': 'george',
                'pid': '-',
                'state': 'stopped',
                'tty': '-',
                'vcpus': '1'
            },
        }
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertEqual(expected, vmctl.status())

    def test_status_single(self):
        '''
        Tests getting status for a single VM.
        '''
        ret = {}
        ret['stdout'] = '   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n' \
                        '    1   123     1    2.9G     150M  ttyp5       ringo web4\n' \
                        '    2     -     1    512M       -       -       george web3\n'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        expected = {
            'web4': {
                'curmem': '150M',
                'id': '1',
                'maxmem': '2.9G',
                'owner': 'ringo',
                'pid': '123',
                'state': 'running',
                'tty': 'ttyp5',
                'vcpus': '1'
            },
        }
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            self.assertEqual(expected, vmctl.status('web4'))

    def test_stop_when_running(self):
        '''
        Tests stopping a VM that is running.
        '''
        ret = {}
        ret['stdout'] = ''
        ret['stderr'] = 'vmctl: sent request to terminate vm 14'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.stop('web1')
            mock_cmd.assert_called_once_with(['vmctl', 'stop', 'web1'],
                output_loglevel='trace', python_shell=False)
            self.assertTrue(res['changes'])

    def test_stop_when_stopped(self):
        '''
        Tests stopping a VM that is already stopped/stopping.
        '''
        ret = {}
        ret['stdout'] = ''
        ret['stderr'] = 'vmctl: terminate vm command failed: Invalid argument'
        ret['retcode'] = 0
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(vmctl.__salt__, {'cmd.run_all': mock_cmd}):
            res = vmctl.stop('web1')
            mock_cmd.assert_called_once_with(['vmctl', 'stop', 'web1'],
                output_loglevel='trace', python_shell=False)
            self.assertFalse(res['changes'])
